/*
L'objet section critique fournit une synchronisation similaire au mutex sauf que ces objets doivent être utilisées par les threads d'un même processus. Mais ce mécanisme est plus performant que les mutex et il est plus simple à utiliser.

Attention :

Si vous ne rendez pas la main à la fin de la section critique, cela bloquera le système. De même, il faut que les sections critiques soient les plus courtes possibles.

Le processus est responsable de l'allocation de la mémoire utilisée par la section critique.
*/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

CRITICAL_SECTION cs; // shared structure

void travail(int i)
{
   int simulationTravail = 1000 + (int)((double)rand() * (i+1000) / (RAND_MAX+1.0));
   printf("la tache%d travaille pendant %.2f s ...\n", i, (float)(simulationTravail/1000.0));
   Sleep(simulationTravail);
   printf("la tache%d a fini son travail ...\n", i);
}

DWORD WINAPI ThreadProc1(LPVOID lpParam)
{
   int n;
   int *pData;
   
   pData = (int*)lpParam;
   n = *pData;
   
   EnterCriticalSection(&cs);
   printf("la tache%d est entre dans une section critique ...\n", n);
   travail(n);
   printf("la tache%d va sortir d'une section critique ...\n", n);
   LeaveCriticalSection(&cs);

   return 0;
}

DWORD WINAPI ThreadProc2(LPVOID lpParam)
{
   int n;
   int *pData;
   
   pData = (int*)lpParam;
   n = *pData;

   EnterCriticalSection(&cs);
   printf("la tache%d est entre dans une section critique ...\n", n);
   travail(n);
   printf("la tache%d va sortir d'une section critique ...\n", n);
   LeaveCriticalSection(&cs);

   return 0;
}

int main(void)
{
   DWORD  dwThreadId1, dwThreadId2;
   HANDLE hThread1, hThread2;
   int n1 = 1, n2 = 2;

   InitializeCriticalSection(&cs);
   
   srand(time(NULL)); //pour la simulation
   
   // Création du thread (autre que le primaire)
   hThread1 = CreateThread(NULL, // default security attributes
                          0, // use default stack size
                          ThreadProc1, // thread function
                          &n1, // argument to thread function
                          0, // use default creation flags
                          &dwThreadId1); // returns the thread identifier

   if (hThread1 == NULL)
   {
      printf( "CreateThread failed (%d).\n", GetLastError());
      ExitProcess(1);
   }
   
   // Création du thread (autre que le primaire)
   hThread2 = CreateThread(NULL, // default security attributes
                          0, // use default stack size
                          ThreadProc2, // thread function
                          &n2, // argument to thread function
                          0, // use default creation flags
                          &dwThreadId2); // returns the thread identifier
   
   if (hThread2 == NULL)
   {
      printf( "CreateThread failed (%d).\n", GetLastError());
      ExitProcess(1);
   }

   // Attente fin des threads
   WaitForSingleObject(hThread1, INFINITE);
   WaitForSingleObject(hThread2, INFINITE);

   CloseHandle(hThread1);
   CloseHandle(hThread2);
   
   DeleteCriticalSection(&cs);
   printf("Fin ...\n");

   return 0;
}
