#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <time.h>

/* Exemple 1 : 1 x Producteur et 1 x Consommateur

Le Producteur ne peut déposer que s'il existe une place libre dans la file.
Le Consommateur ne peut prélever une requête que si elle a été préalablement déposée.
La file est une ressource épuisable
La file n'est pas une ressource critique car le Producteur et le Consommateur n'en manipulent pas la même partie.

2 sémaphores :
-> La synchronisation du Producteur vis à vis de la file est à mettre en place
-> La synchronisation du Consommateur vis à vis du Producteur est à mettre en place

*/

/* taille de la file */
#define L 5

/* la file */
int f[L];
int in = 0, out = 0;

/* Les semaphores */
HANDLE nbRequetes; // init 0
HANDLE nbCasesVides; //  init L

void P(HANDLE s)
{
   WaitForSingleObject(s, INFINITE);
}

void V(HANDLE s)
{
   ReleaseSemaphore(s, 1, NULL);
}

int construireRequete(int n)
{
   int simulationTravail = 100 + (int)((double)rand() * (n+1000) / (RAND_MAX+1.0));
   int requete = 1 + (int)((double)rand() * (n+100) / (RAND_MAX+1.0));
   //printf("la tache%d construit une requete en %.3f s ...\n", n, (float)(simulationTravail/1000.0));
   Sleep(simulationTravail);
   
   return requete;
}

void deposerRequete(int n, int requete)
{
   printf("la tache%d producteur depose la requete %d dans la file a la position %d ...\n", n, requete, in);
   f[in] = requete;
   in = (in + 1) % L;   
}

int retirerRequete(int n)
{
   int requete;
   
   printf("la tache%d consommateur retire une requete de la file a la position %d ...\n", n, out);
   requete = f[out];
   f[out] = 0; /* élement vide */
   out = (out + 1) % L;
   
   return requete;
}

void traiterRequete(int n, int requete)
{
   int simulationTravail = 1000 + (int)((double)rand() * (n+1000) / (RAND_MAX+1.0));
   //printf("la tache%d traite la requete en %.2f s ...\n", n, (float)(simulationTravail/1000.0));
   Sleep(simulationTravail*2); // on retarde un peu le consommateur !
   requete = 0; // inutilement inutile !
}

DWORD WINAPI producteur(LPVOID nt)
{
   int n = *(int *)nt; // numéro de tache
   int requete;
   int i;
   
   // on produit pendant un certain temps ...
   for(i=0;i<(L*3);i++)
   {
      requete = construireRequete(n);
      
      //printf("la tache%d producteur attend une place libre ...\n", n);
      P(nbCasesVides);     
      
      deposerRequete(n, requete);
      
      V(nbRequetes);
      //printf("la tache%d producteur signale qu'une requete est disponible dans la file ...\n", n);
   }
   return 0;
}

DWORD WINAPI consommateur(LPVOID nt)
{
   int n = *(int *)nt; // numéro de tache
   int requete;
   int i;
   
   //on consomme pendant un certain temps ...
   for(i=0;i<(L*2);i++)
   {   
      //printf("la tache%d consommateur attend une requete disponible dans la file ...\n", n);
      P(nbRequetes);
      
      requete = retirerRequete(n);
      
      V(nbCasesVides);

      printf("la tache%d consommateur va traiter la requete %d et libere la position ...\n", n, requete);
      traiterRequete(n, requete);
   }
   return 0;
}

int main(void)
{
   /* les threads */
   DWORD  dwThreadId1, dwThreadId2;
   HANDLE hThread1, hThread2;
   int t1 = 1, t2 = 2; // numéros des taches

   /* init des semaphores */
   nbRequetes   = CreateSemaphore(NULL, 0, L, (LPCTSTR)"SemaphoreNbRequetes");
   nbCasesVides = CreateSemaphore(NULL, L, L, (LPCTSTR)"SemaphoreNbCasesVides");
   
   // si dans un autre processus alors ouverture d'un sémaphore déjà créé
   //nbRequetes   = OpenSemaphore(SEMAPHORE_ALL_ACCESS, FALSE, (LPCTSTR)"SemaphoreNbRequetes");
   //nbCasesVides = OpenSemaphore(SEMAPHORE_ALL_ACCESS, FALSE, (LPCTSTR)"SemaphoreNbCasesVides");

   srand(time(NULL)); //pour la simulation

   /* demarre les threads */
   hThread1 = CreateThread(NULL, // default security attributes
                           0, // use default stack size
                           consommateur, // thread function
                           &t1, // argument to thread function
                           0, // use default creation flags
                           &dwThreadId1); // returns the thread identifier
   hThread2 = CreateThread(NULL, // default security attributes
                           0, // use default stack size
                           producteur, // thread function
                           &t2, // argument to thread function
                           0, // use default creation flags
                           &dwThreadId2); // returns the thread identifier

   /* attends la fin des threads */
   WaitForSingleObject(hThread1, INFINITE);
   WaitForSingleObject(hThread2, INFINITE);
   CloseHandle(hThread1);
   CloseHandle(hThread2);
   CloseHandle(nbRequetes);
   CloseHandle(nbCasesVides);
   printf("Fin ...\n");
   return 0;
}
